﻿using PaintDotNet;
using PaintDotNet.Effects;
using PaintDotNet.IndirectUI;
using PaintDotNet.PropertySystem;
using PaintDotNet.Rendering;
using System;
using System.Collections.Generic;
using System.Drawing;
//using System.Drawing.Drawing2D;//added for linearGradientBrushes
using System.Threading.Tasks; // .NET 4.0 required
using System.Reflection;
using System.Runtime.InteropServices;

[assembly: AssemblyTitle("EdgeShader")]
[assembly: AssemblyDescription("Detects object edges only, stores in an ArrayList, colours pixels depending on distance from an edge")]
[assembly: AssemblyConfiguration("edges|shading|outlines")]
[assembly: AssemblyCompany("Red ochre (John Robbins)")]
[assembly: AssemblyProduct("EdgeShader")]
[assembly: AssemblyCopyright("Copyright © 2015 Red ochre (John Robbins)")]
[assembly: AssemblyTrademark("")]
[assembly: AssemblyCulture("")]
[assembly: ComVisible(false)]
[assembly: AssemblyVersion("1.0.2")]

namespace EdgeShader
{
	public class PluginSupportInfo : IPluginSupportInfo
	{
		public string Author
		{
			get
			{
				return "Red ochre (John Robbins)";
			}
		}
		public string Copyright
		{
			get
			{
				return ((AssemblyCopyrightAttribute)base.GetType().Assembly.GetCustomAttributes(typeof(AssemblyCopyrightAttribute), false)[0]).Copyright;
			}
		}

		public string DisplayName
		{
			get
			{
				return ((AssemblyProductAttribute)base.GetType().Assembly.GetCustomAttributes(typeof(AssemblyProductAttribute), false)[0]).Product;
			}
		}

		public Version Version
		{
			get
			{
				return base.GetType().Assembly.GetName().Version;
			}
		}

		public Uri WebsiteUri
		{
			get
			{
				return new Uri("http://www.getpaint.net/redirect/plugins.html");
			}
		}
	}

	[PluginSupportInfo(typeof(PluginSupportInfo), DisplayName = "EdgeShader")]
	public class EdgeShaderPlugin : PropertyBasedEffect
	{
        public static string StaticName
        {
            get
            {
                return "EdgeShader";
            }
        }

        public static Image StaticIcon
        {
            get
            {
                return EdgeShader.Properties.Resources.EffectIcon;
            }
        }

        public EdgeShaderPlugin()
            : base(StaticName, StaticIcon, "Test", EffectFlags.SingleRenderCall) // Force Single Render Call
        {
           
        }

        public enum PropertyNames
        {
            OpacityThresh,
            Maxdist,
            EffectMix,
            SwapColours
        }
     
        protected override PropertyCollection OnCreatePropertyCollection()
        {
            List<Property> props = new List<Property>();
            props.Add(new Int32Property(PropertyNames.OpacityThresh, 1, 1, 255));
            props.Add(new Int32Property(PropertyNames.Maxdist, 100, 1, 500));
            props.Add(new DoubleProperty(PropertyNames.EffectMix, 100, 0, 100));
            props.Add(new BooleanProperty(PropertyNames.SwapColours, false));
          
            return new PropertyCollection(props);
        }

        protected override ControlInfo OnCreateConfigUI(PropertyCollection props)
        {
            ControlInfo configUI = CreateDefaultConfigUI(props);
            configUI.SetPropertyControlValue(PropertyNames.OpacityThresh, ControlInfoPropertyNames.DisplayName, "Opacity threshold");
            configUI.SetPropertyControlValue(PropertyNames.Maxdist, ControlInfoPropertyNames.DisplayName, "Maximum distance");
            configUI.SetPropertyControlValue(PropertyNames.EffectMix, ControlInfoPropertyNames.DisplayName, "Effect mix %");
            configUI.SetPropertyControlValue(PropertyNames.EffectMix, ControlInfoPropertyNames.SliderLargeChange, 0.25);
            configUI.SetPropertyControlValue(PropertyNames.EffectMix, ControlInfoPropertyNames.SliderSmallChange, 0.05);
            configUI.SetPropertyControlValue(PropertyNames.EffectMix, ControlInfoPropertyNames.UpDownIncrement, 0.01);
            configUI.SetPropertyControlValue(PropertyNames.SwapColours, ControlInfoPropertyNames.DisplayName, string.Empty);
            configUI.SetPropertyControlValue(PropertyNames.SwapColours, ControlInfoPropertyNames.Description, "Swap colours");
         return configUI;
        }

        protected override void OnSetRenderInfo(PropertyBasedEffectConfigToken newToken, RenderArgs dstArgs, RenderArgs srcArgs)
        {
            this.threshA = newToken.GetProperty<Int32Property>(PropertyNames.OpacityThresh).Value;
            this.maxdistance = newToken.GetProperty<Int32Property>(PropertyNames.Maxdist).Value;
            this.effectmix = newToken.GetProperty<DoubleProperty>(PropertyNames.EffectMix).Value;
            this.swapcols = newToken.GetProperty<BooleanProperty>(PropertyNames.SwapColours).Value;
            base.OnSetRenderInfo(newToken, dstArgs, srcArgs);
        }



		protected override void OnCustomizeConfigUIWindowProperties(PropertyCollection props)
		{
			// Change the effect's window title
			props[ControlInfoPropertyNames.WindowTitle].Value = "EdgeShader (working title only) 15/8/2015";
			base.OnCustomizeConfigUIWindowProperties(props);
		}

		#region User Entered Code
		/* =================================================== */
		/* EdgeShader    */
		/* Name    */
		/* (c) 2015 Red ochre (John Robbins)     */
		/* comment   */
		/* Description: uses distance to nearest edge to colour an image*/
		/*     */
		/* ========================================== ======== */

		// Name: EdgeShader
		// Author: Red ochre (John Robbins)
		// Submenu: Advanced
		// URL: http://www.getpaint.net/redirect/plugins.html
		// Title:EdgeShader                       Red ochre Feb 2015

		#region UICode
		int threshA = 255; // [0,255] Opacity threshold
		int maxdistance = 50; // [0,100] Maximum distance
		double effectmix = 50; // [0,100] Effect mix %
		bool swapcols = false;
		#endregion
		//BASIC IDEA: 
		//load up arraylists with edges
		//then do normal y/x loops for every opaque pixel and loop through arrays to find nearest edge pixel distance
		//change pixel colour according to distance and mix with src
		//- will be slow but slightly quicker than checking every surrounding pixel - quicker for less edge and more distance
		//slower for complex edges and short distances - comparing with a blurred surface is better for that.
		//did have a version of this working finding colour/tone edges too but far too slow using this method.
		//I have ideas for another (more efficient) way to do this.

		//PROBLEM - am I using pointers correctly in y/x loops at end?
		//*srcpix works but *destpix crashes Pdn although fine in VS? - hence removed and using dest[x,y]
		//NOTE: I am using another surface 'dest' as multithread detection doesn't know when it's found an edge - I assume.


		Boolean AlphaEdges(Surface src, int x, int y)
		{
			bool Aedge = false;

			ColorBgra TL = src[x - 1, y - 1];//Top Left
			ColorBgra TM = src[x, y - 1];//Top Middle
			ColorBgra TR = src[x + 1, y - 1];//Top Right
			ColorBgra ML = src[x - 1, y];//Middle Left
			//ColorBgra MM = src[x, y];//Middle or 'current pixel'- not required here
			ColorBgra MR = src[x + 1, y];//Middle Right
			ColorBgra BL = src[x - 1, y + 1];//Bottom Left
			ColorBgra BM = src[x, y + 1];// Bottom Middle
			ColorBgra BR = src[x + 1, y + 1];//Bottom Right
			if (TL.A < threshA || TM.A < threshA || TR.A < threshA || ML.A < threshA || MR.A < threshA || BL.A < threshA || BM.A < threshA || BR.A < threshA) { Aedge = true; }
			return Aedge;
		}

		protected override void OnRender(Rectangle[] renderRects, int startIndex, int length)
		{
			Surface src = this.SrcArgs.Surface;
			Surface dst = this.DstArgs.Surface;

			List<PointInt32> edges = new List<PointInt32>();

			// Get Edges Loop
			Parallel.For(0, renderRects.Length, (i) => // Parallel.For is in System.Threading.Tasks Namespace --> .NET 4.0 required
			{
				if (this.IsCancelRequested)
					return;
				Rectangle rect = renderRects[i];

				Parallel.For(rect.Top+1, rect.Bottom-1, (y) =>
				{
					if (this.IsCancelRequested)
						return;

					List<PointInt32> tempEdges = new List<PointInt32>(rect.Width);

					for (int x = rect.Left+1; x < rect.Right-1; x++)
					{
						ColorBgra pixel = src[x, y];
						if (pixel.A > this.threshA && AlphaEdges(src, x, y))
							tempEdges.Add(new PointInt32(x, y));
					}

					lock(edges)
					{
						edges.AddRange(tempEdges);
					}
				});
			});

			// Render Loop
			Parallel.For(0, renderRects.Length, (i) =>
			{
				if (this.IsCancelRequested)
					return;
				Rectangle rect = renderRects[i];

				Parallel.For(rect.Top, rect.Bottom, (y) =>
				{
					if (this.IsCancelRequested)
						return;

					for (int x = rect.Left; x < rect.Right; x++)
					{
						int minDistSQ = int.MaxValue;
						foreach (PointInt32 edge in edges)
						{
							// calc min distance
						}

						// calc pixel based on minDistSQ
						ColorBgra pixel = new ColorBgra();
						// ...
						dst[x,y] = pixel;
					}
				});
			});
		}

		#endregion
	}
}
